'use strict';

var _promise = require('babel-runtime/core-js/promise');

var _promise2 = _interopRequireDefault(_promise);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var fs = require('fs');
var spawn = require('child_process').spawn;
var reporter = require('./reporter.js');

var prepend = '(function(process, require, console) {\n';
var append = '\n})'; // dont remove \n

var sentinel = function () {
  var b = new Buffer(12);
  b.writeInt32LE(0x26e0c928, 0);
  b.writeInt32LE(0x41f32b66, 4);
  b.writeInt32LE(0x3ea13ccf, 8);
  return b;
}();

var padding = function padding(sz) {
  return new Buffer({
    0: 0, 1: 3, 2: 2, 3: 1
  }[sz % 4]);
};

var buffer4 = function buffer4(v) {
  var b = new Buffer(4);
  b.writeInt32LE(v, 0);
  return b;
};

function producer(opts, cb) {

  var stripe = opts.stripe.slice(0);
  stripe.unshift(prepend);
  stripe.push(append);
  var extra = new Buffer(stripe.join(''));

  var fabricator = void 0;
  var fabricatorName = opts.fabricatorName;
  var fabricatorArgs = ['--fabricator'];

  try {
    fabricator = spawn(fabricatorName, fabricatorArgs);
    fabricator.stdin.write(buffer4(extra.length));
    fabricator.stdin.write(extra);
    fabricator.stdin.end();
  } catch (error) {
    reporter.report(fabricatorName, 'error', ['Failed to spawn fabricator ' + fabricatorName, error.toString()], error);
    return cb(error);
  }

  var snapshotChunks = [];

  fabricator.stderr.on('data', function (ch) {
    process.stdout.write(ch);
  });

  fabricator.stderr.on('end', fabricatorMaybeExit);

  fabricator.stdout.on('data', function (ch) {
    snapshotChunks.push(ch);
  });

  fabricator.stdout.on('end', fabricatorMaybeExit);

  fabricator.on('exit', function (code, signal) {
    if (code !== 0) {
      var error = new Error('Fabricator exited with code/signal ' + (code || signal));
      reporter.report('', 'error', error.message, error);
      return cb(error);
    }
    fabricatorMaybeExit();
  });

  var fabricatorCounter = 0;
  function fabricatorMaybeExit() {

    fabricatorCounter += 1;
    if (fabricatorCounter !== 3) return;

    var snapshot = Buffer.concat(snapshotChunks);

    if (snapshot.length === 0) {
      var error = new Error('Empty fabricator output');
      reporter.report('', 'error', error.message, error);
      return cb(error);
    }

    var houseName = opts.houseName || fabricatorName;
    var house = fs.readFileSync(houseName);

    cb(null, Buffer.concat([house, padding(house.length), sentinel, buffer4(snapshot.length), snapshot]));
  }
}

module.exports = function (opts) {
  return new _promise2.default(function (resolve, reject) {
    producer(opts, function (error, bundle) {
      if (error) return reject(error);
      resolve(bundle);
    });
  });
};